/***************************************************************************
 *                                                                         *
 *   MAILCRAWL                                                             *
 *   =========                                                             *
 *                                                                         *
 *   Copyright (C)                                                         *
 *   Juan Pedro Bolivar Puente                                             *
 *   Francisco Manuel Herrero Pérez                                        *
 *   2007                                                                  *
 *                                                                         *
 *   This program is free software: you can redistribute it and/or modify  *
 *   it under the terms of the GNU General Public License as published by  *
 *   the Free Software Foundation, either version 3 of the License, or     *
 *   (at your option) any later version.                                   *
 *                                                                         *
 *   This program is distributed in the hope that it will be useful,       *
 *   but WITHOUT ANY WARRANTY; without even the implied warranty of        *
 *   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the         *
 *   GNU General Public License for more details.                          *
 *                                                                         *
 *   You should have received a copy of the GNU General Public License     *
 *   along with this program.  If not, see <http://www.gnu.org/licenses/>. *
 *                                                                         *
 ***************************************************************************/

/*
  TODO: error handling.
 */

#include <config.h>
#include <string.h>
#include <unistd.h>
#include <signal.h>

#include "log.h"
#include "common.h"
#include "mailcrawl.h"

#define HYSTERESIS_FACTOR 0.5
#define CACHE_FACTOR 0.5
#define DEFAULT_MAX_CACHE (1<<18)
#define DEFAULT_MAX_PAGES -1
#define TMP_FILE_PATH     "/tmp/"
#define BUF_SIZE 1024

static volatile sig_atomic_t _must_kill = false;

struct qelem_s
{
  char* url;
  int type;
};

typedef struct qelem_s qelem_t;
typedef qelem_t* qelem_p;

static qelem_p
qelem_create(char* url, int type)
{
  qelem_p e;

  e = malloc(sizeof(qelem_t));
  if (e)
    {
      e->url = url;
      e->type = type;
    }
  
  return e;
}

static char*
mailcrawl_tmp_file_name (mailcrawl_p mc)
{
  static int file_id = 0;
  int my_file_id;
  char* tmp_file_name;

  pthread_mutex_lock (&mc->mutex);
  my_file_id = file_id ++;
  pthread_mutex_unlock (&mc->mutex);
  
  tmp_file_name = malloc (sizeof(char) *
			  (strlen(TMP_FILE_PATH) +
			   strlen(PACKAGE) +
			   32));
  strcpy (tmp_file_name, TMP_FILE_PATH);
  strcat (tmp_file_name, PACKAGE);
  strcat (tmp_file_name, ntoa (getpid()));
  strcat (tmp_file_name, ntoa (my_file_id));

  return tmp_file_name;
}

void
mailcrawl_kill_handler()
{
  if (!_must_kill)
    {
      _must_kill = true;
      fprintf(stderr, "Saliendo...\n");
    }
}

mailcrawl_p
mailcrawl_create ()
{
  mailcrawl_p mc;

  mc = malloc (sizeof(mailcrawl_t));

  mc->output = stdout;
    
  mc->max_cache = DEFAULT_MAX_CACHE; 
  mc->max_pages = DEFAULT_MAX_PAGES;
  
  mc->next_pages = queue_create ((queue_dispose_fn)&free);
  
  mc->found_pages = cache_create ();
  mc->found_mails = cache_create ();

  mc->decreasing = false;

  pthread_mutex_init (&mc->mutex, NULL);  
  return mc;
}

void
mailcrawl_free (mailcrawl_p mc)
{  
  queue_free (mc->next_pages);

  cache_free (mc->found_pages);
  cache_free (mc->found_mails);

  pthread_mutex_destroy (&mc->mutex);
  
  free(mc);
}

bool
mailcrawl_next(mailcrawl_p mc, char** url_base_out, char** url_out)
{
  int i;

  int type;
  char* url;
  char* url_base;

  pthread_mutex_lock (&mc->mutex);

  if (queue_empty (mc->next_pages))
    {
      pthread_mutex_unlock (&mc->mutex);
      return true;
    }
  
  type = ((qelem_p) queue_front (mc->next_pages))->type;
  url  = ((qelem_p) queue_front (mc->next_pages))->url;
  
  url_base = malloc (sizeof(char) * (strlen(url) + 2));
  strcpy (url_base, url);

  if (type == URL_FILE)
    {
      for (i = strlen (url_base) - 1; i > 0 && url_base[i] != '/'; --i);
      url_base[i+1] = '\0';
    }
  else if (type == URL_DIR)
    {
      for (i = strlen(url_base) - 1;
	   i > 1 && url_base[i] != '?';
	   --i);

      if (url_base[i] == '?')
	{
	  url_base[i] = '/';
	  url_base[i+1] = '\0';
	}
      else if (url_base[ strlen(url_base) - 1 ] != '/')
	{
	  url_base[strlen(url_base)+1] = '\0';
	  url_base[strlen(url_base)] = '/';
	}
    }

  queue_pop (mc->next_pages);

  *url_base_out = url_base;
  *url_out = url;

  pthread_mutex_unlock (&mc->mutex);

  return false;
}

bool 
mailcrawl_fetch (mailcrawl_p mc,
		 const char* file,
		 const char* url)
{
  FILE* fout;
  CURL* curl;
  CURLcode ret;

  log_msg (LOG_INFO, "Visitando URL: %s.", url);
  
  fout = fopen (file, "w");

  if (!fout)
    {
      log_msg (LOG_ERROR,
	       "No se pudo abrir para escritura el fichero %s",
	       file);
      return ERROR;
    }

  curl = curl_easy_init();
  
  curl_easy_setopt (curl, CURLOPT_WRITEDATA, fout);
  curl_easy_setopt (curl, CURLOPT_URL, url);

  ret = curl_easy_perform (curl);

  curl_easy_cleanup (curl);
  
  fclose(fout);

  if (ret != CURLE_OK)
    {
      log_msg(LOG_WARNING,
	      "No se pudo descargar %s: %s",
	      url, curl_easy_strerror (ret));
      return ERROR;
    }
  
  return OK;
}

static void
mailcrawl_iteration (void* arg)
{
  mailcrawl_p mc = arg;
  char* url_base = NULL;
  char* url = NULL;
  char* tmp_file_name = NULL;

  if (mailcrawl_next (mc, &url_base, &url))
    return;
  
  tmp_file_name = mailcrawl_tmp_file_name (mc);
  if (mailcrawl_fetch (mc, tmp_file_name, url) == OK && !_must_kill)
    mailcrawl_lex (mc, tmp_file_name, url_base);

  remove (tmp_file_name);
  free (tmp_file_name);
  free (url_base);
  free (url);
  
  pthread_mutex_lock (&mc->mutex);
  if (queue_empty (mc->next_pages) &&
      thread_pool_num_working (mc->thread_pool) < 2)
    {
      _must_kill = true; /* FIXME */
    }
  pthread_mutex_unlock (&mc->mutex);
}

void
mailcrawl_run (mailcrawl_p mc, size_t num_threads)
{
  thread_pool_p thread_pool;
  int max = mc->max_pages;

  thread_pool = thread_pool_create (num_threads);
  mc->thread_pool = thread_pool;
  
  while (max && !_must_kill)
    {
      thread_pool_execute (thread_pool,
			   mailcrawl_iteration,
			   mc);
      if (max > 0)
	max--;
    }
  
  thread_pool_free (mc->thread_pool);
  mc->thread_pool = NULL;
}

bool
mailcrawl_load_resume_file(mailcrawl_p mc, const char* file)
{
  return mailcrawl_lex(mc, file, "");
#if 0
  FILE* fin = fopen(file, "r");
  
  char buf[BUF_SIZE];
  int type;
  
  if (fin)
    {
      log_msg(LOG_INFO, "Cargando fichero de reset %s.", file);
      
      while(fgets(buf, BUF_SIZE, fin))
	{
	  fscanf("%d\n", &type);
	  buf[strlen(buf)-1] = '\0';
	  mailcrawl_add_url(mc, buf, type);
	}
      
      fclose(fin);

      return OK;
    }
  else
    {
      log_msg(LOG_INFO,
	      "No se pudo abrir el fichero de resume %s para lectura.", file);

      return ERROR;
    }
#endif
}

bool
mailcrawl_save_resume_file(mailcrawl_p mc, const char* file)
{
  FILE* fout = fopen(file, "w");
  qelem_p e;

  if (fout)
    {
      log_msg(LOG_INFO, "Escribiendo fichero de reset %s.", file);

      pthread_mutex_lock (&mc->mutex);
      int i = queue_size (mc->next_pages);
      while(i--)
	{
	  e = queue_front (mc->next_pages);
	  fprintf (fout, "\"%s\"\n", e->url);
	  queue_push (mc->next_pages, qelem_create (e->url, e->type));
	  queue_pop (mc->next_pages);
	}
      pthread_mutex_unlock (&mc->mutex);
      
      fclose(fout);

      return OK;
    }
  else
    {
      log_msg(LOG_INFO,
	      "No se pudo abrir el fichero de resume %s para escritura.", file);

      return ERROR;
    }
}

INLINE bool
mailcrawl_queue_full(mailcrawl_p mc)
{
  return (queue_size(mc->next_pages) >
	  (mc->decreasing ?
	   mc->max_cache * HYSTERESIS_FACTOR :
	   mc->max_cache));
}

bool
mailcrawl_put_email(mailcrawl_p mc, const char* email)
{ 
  pthread_mutex_lock (&mc->mutex);

  if (cache_add(mc->found_mails, email))
    {
      if (cache_size(mc->found_mails) > mc->max_cache)
	cache_adjust(mc->found_mails, mc->max_cache * CACHE_FACTOR);
	
      fprintf(mc->output, "%s\n", email);
      fflush(mc->output);
      
      log_msg(LOG_TRIVIAL, "Email añadido %s.", email);
      pthread_mutex_unlock (&mc->mutex);
      return true;
    }
    
  pthread_mutex_unlock (&mc->mutex);
  
  return false;
}

INLINE int
mailcrawl_add_my_url(mailcrawl_p mc, char* cpy, int url_type)
{
  /* FIXME: Leak on else? */
  pthread_mutex_lock (&mc->mutex);
  if (!mailcrawl_queue_full(mc))
    {
      if (cache_add(mc->found_pages, cpy))
	{
	  if (cache_size(mc->found_pages) > mc->max_cache)
	    cache_adjust(mc->found_pages, mc->max_cache * CACHE_FACTOR);
	  
	  queue_push(mc->next_pages, qelem_create(cpy, url_type));
	  	  
	  log_msg(LOG_TRIVIAL, "Enlace añadido %s.", cpy);
	  pthread_mutex_unlock (&mc->mutex);
	  return true;
	}
    }

  mc->decreasing = true;
  free (cpy);

  pthread_mutex_unlock (&mc->mutex);
  
  return false;
}

bool
mailcrawl_add_url(mailcrawl_p mc, const char* url, int url_type)
{
  char* cpy;
  
  cpy = malloc(sizeof(char) * (strlen(url) + 1));
  strcpy(cpy, url);

  return mailcrawl_add_my_url(mc, cpy, url_type);
}

void
mailcrawl_load_urls (mailcrawl_p mc, const char** urls, int n_urls)
{
  char* tmp_file_name = mailcrawl_tmp_file_name (mc); 
  FILE* fout = fopen (tmp_file_name, "w");

  if (!fout)
    {
      log_msg(LOG_FATAL,
	      "No se pudo abrir el fichero temporal %s.",
	      tmp_file_name);
      exit(-1); /* FIXME */
    }

  while(n_urls--)
    fprintf(fout, "\"%s\"\n", *urls++);
  
  fclose(fout);

  mailcrawl_lex (mc, tmp_file_name, "");
  free (tmp_file_name);
}

bool
mailcrawl_add_rel(mailcrawl_p mc,
		  const char* url_base,
		  const char* rel,
		  int url_type)
{
  char* cpy;
    
  cpy = malloc (sizeof(char) * (strlen(rel) + strlen(url_base) + 1));
  strcpy (cpy, url_base);
  strcat (cpy, rel);

  return mailcrawl_add_my_url (mc, cpy, url_type);
}
