/***************************************************************************
 *                                                                         *
 *   MAILCRAWL                                                             *
 *   =========                                                             *
 *                                                                         *
 *   Copyright (C)                                                         *
 *   Juan Pedro Bolivar Puente                                             *
 *   Francisco Manuel Herrero Pérez                                        *
 *   2007                                                                  *
 *                                                                         *
 *   This program is free software: you can redistribute it and/or modify  *
 *   it under the terms of the GNU General Public License as published by  *
 *   the Free Software Foundation, either version 3 of the License, or     *
 *   (at your option) any later version.                                   *
 *                                                                         *
 *   This program is distributed in the hope that it will be useful,       *
 *   but WITHOUT ANY WARRANTY; without even the implied warranty of        *
 *   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the         *
 *   GNU General Public License for more details.                          *
 *                                                                         *
 *   You should have received a copy of the GNU General Public License     *
 *   along with this program.  If not, see <http://www.gnu.org/licenses/>. *
 *                                                                         *
 ***************************************************************************/

#include <libgen.h>
#include <getopt.h>
#include <stdio.h>
#include <signal.h>

#include "log.h"
#include "mailcrawl.h"

static int
parse_args(mailcrawl_p mc,
	   FILE** log_file,
	   const char** reset_file,
	   size_t* num_threads,
	   int argc, char* argv[]);

static void
show_help(const char* name);

static void
show_version();


int main(int argc, char* argv[])
{
  size_t num_threads = 1;
  mailcrawl_p mc;
  FILE* log_file = NULL;
  const char* reset_file = NULL;
  
  log_set_level(LOG_WARNING);
 
  signal(SIGINT, &mailcrawl_kill_handler);
  signal(SIGTERM, &mailcrawl_kill_handler);
  
  mc = mailcrawl_create();
  
  if (parse_args(mc, &log_file, &reset_file, &num_threads, argc, argv) == false)
    {
      if (reset_file)
	mailcrawl_load_resume_file(mc, reset_file);
      mailcrawl_run(mc, num_threads);
    }
  
  if (mailcrawl_get_output(mc) != stdout)
    fclose(mailcrawl_get_output(mc));

  if (reset_file)
    mailcrawl_save_resume_file(mc, reset_file);


  mailcrawl_free(mc);
  
  if (log_file)
    fclose(log_file);
  
  exit(0);
}

void
show_help(const char* name)
{
  show_version();
  
  printf("\n"
	 "Uso:\n"
	 "  %s [ opciones ] url_inicial [ url_iniciales ]\n"
	 "\n"
	 "Opciones:\n"
	 "  -h, --help             Muestra esta ayuda.\n"
	 "  --version              Muestra la versión del programa.\n"
	 "  -v, --verbose[=valor]  Elige la cantidad de información a mostrar en pantalla\n"
	 "                         entre %d, información más abundante y %d, ningún mensaje.\n"
	 "  -l, --log <fichero>    Guarda el registro en un fichero.\n"
	 "  -m, --max <valor>      Elige un número máximo de páginas a visitar. Si es\n"
	 "                         negativo equivale a infinito (limitado por la memoria\n"
	 "                         del ordenador). Por defecto es -1.\n"
	 "  -o, --output <fichero> Elige el fichero de salida. Si no se especifica se\n"
	 "                         mostrará en la salida estándar.\n"
	 "  -r, --resume <fichero> Elige un fichero de dónde cargar la cola de páginas\n"
	 "                         a visitar, y dónde guardar la cola a actual tras\n"
	 "                         finalizar la ejecución del programa\n."
	 "  -c, --cache <valor>    Elige un valor máximo para el número de páginas a\n"
	 "                         a guardar en la cola y en la caché de páginas y mails\n"
         "                         ya encontrados.\n"
	 "  -t, --threads <valor>  Elige el numero de hebras con las que buscar.\n",
	 name, LOG_TRIVIAL, LOG_FATAL);
}

void
show_version()
{
  printf("%s version %s\n", PACKAGE, VERSION);
}

enum
  {
    OPT_HELP    = 'h',
    OPT_VERBOSE = 'v',
    OPT_LOG     = 'l',
    OPT_MAX     = 'm',
    OPT_OUTPUT  = 'o',
    OPT_CACHE   = 'c',
    OPT_RESET   = 'r',
    OPT_THREADS = 't',
    OPT_VERSION,
    N_OPTS
  };

int
parse_args(mailcrawl_p mc,
	   FILE** log_file,
	   const char** reset_file,
	   size_t* num_threads,
	   int argc, char** argv)
{
  static struct option long_ops[] =
    {
      {"help", no_argument, NULL, OPT_HELP},
      {"version", no_argument, NULL, OPT_VERSION},
      {"verbose", optional_argument, NULL, OPT_VERBOSE},
      {"max", required_argument, NULL, OPT_MAX},
      {"log", required_argument, NULL, OPT_LOG},
      {"output", required_argument, NULL, OPT_OUTPUT},
      {"resume", required_argument, NULL, OPT_RESET},
      {"cache", required_argument, NULL, OPT_CACHE},
      {"threads", optional_argument, NULL, OPT_THREADS},
      {NULL, 0, NULL, 0}
    };

  int opt;
  int opt_index;
  FILE* out_file;
  
  while(1)
    {
      opt = getopt_long (argc, argv, "hl:m:o:c:r:t:v::",
			 long_ops, &opt_index);

      if (opt == -1)
	break;

      switch(opt)
	{ 
	case OPT_VERSION:
	  show_version();
	  return true;
	  break;
	  
	case OPT_HELP:
	  show_help(basename(argv[0]));
	  return true;
	  break;

	case OPT_VERBOSE:
	  log_set_level(optarg ? atoi(optarg) : LOG_INFO);
	  break;

	case OPT_LOG:
	  *log_file = fopen(optarg, "a");
	  if (!*log_file)
	    log_msg(LOG_ERROR, "No se pudo abrir el fichero de registro %s.", optarg);
	  else
	    log_set_file(*log_file);
	  break;

	case OPT_MAX:
	  mailcrawl_set_max_pages(mc, atoi(optarg));
	  break;

	case OPT_OUTPUT:
	  out_file = fopen(optarg, "a");
	  if (!out_file)
	    {
	      log_msg(LOG_FATAL, "No se pudo abrir el fichero de salida %s.", optarg);
	      return true;
	    }
	  mailcrawl_set_output(mc, out_file);
	  break;

	case OPT_CACHE:
	  mailcrawl_set_max_cache(mc, atoi(optarg));
	  break;
	  
	case OPT_RESET:
	  *reset_file = optarg;
	  break;

	case OPT_THREADS:
	  *num_threads = atoi (optarg);
	  break;
	  
	default:
	  break;
	}
    }

  if (optind == argc)
    {
      fprintf(stderr, "Debe pasar al menos una URL inicial.\n");
      return true;
    }

  mailcrawl_load_urls(mc, (const char**)argv + optind, argc - optind);
  
  return false;
}
