/***************************************************************************
 *                                                                         *
 *   MAILCRAWL                                                             *
 *   =========                                                             *
 *                                                                         *
 *   Copyright (C)                                                         *
 *   Juan Pedro Bolivar Puente                                             *
 *   Francisco Manuel Herrero Pérez                                        *
 *   2009                                                                  *
 *                                                                         *
 *   This program is free software: you can redistribute it and/or modify  *
 *   it under the terms of the GNU General Public License as published by  *
 *   the Free Software Foundation, either version 3 of the License, or     *
 *   (at your option) any later version.                                   *
 *                                                                         *
 *   This program is distributed in the hope that it will be useful,       *
 *   but WITHOUT ANY WARRANTY; without even the implied warranty of        *
 *   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the         *
 *   GNU General Public License for more details.                          *
 *                                                                         *
 *   You should have received a copy of the GNU General Public License     *
 *   along with this program.  If not, see <http://www.gnu.org/licenses/>. *
 *                                                                         *
 ***************************************************************************/

#include <stdlib.h>
#include "thread-pool.h"

enum thread_pool_worker_status_e
  {
    THREAD_POOL_WORKER_IDLE,
    THREAD_POOL_WORKER_WORKING,
    THREAD_POOL_WORKER_STOP,
    THREAD_POOL_WORKER_ERROR
  };
typedef enum thread_pool_worker_status_e thread_pool_worker_status_t;

struct thread_pool_worker_s
{
  thread_pool_p pool;
  thread_pool_worker_status_t status;
  void (*task_func) (void*);
  void* task_data;
  
  pthread_t thread;
  pthread_mutex_t mutex;
  pthread_cond_t cond;
};
typedef struct thread_pool_worker_s thread_pool_worker_t;
typedef thread_pool_worker_t* thread_pool_worker_p;

struct thread_pool_s
{
  thread_pool_worker_t* workers;
  size_t num_workers;
  size_t num_working;
  
  pthread_mutex_t mutex;
  pthread_cond_t cond;
};

size_t
thread_pool_num_working (thread_pool_p tp)
{
  return tp->num_working;
}

static void*
thread_pool_worker_function (void* arg)
{
  thread_pool_worker_p w = arg;

  pthread_mutex_lock (&w->mutex);
  while (1)
    {
      while (w->status == THREAD_POOL_WORKER_IDLE)
	{
	  pthread_cond_wait (&w->cond, &w->mutex);
	}
      if (w->status == THREAD_POOL_WORKER_STOP)
	break;
      pthread_mutex_unlock (&w->mutex);

      w->pool->num_working ++;
      w->task_func (w->task_data);
      w->pool->num_working --;
      
      pthread_mutex_lock (&w->mutex);
      if (w->status != THREAD_POOL_WORKER_STOP)
	{
	  w->status = THREAD_POOL_WORKER_IDLE;
	}
      pthread_cond_signal (&w->pool->cond);
    }
  pthread_mutex_unlock (&w->mutex);
  
  return NULL;
}

static void
thread_pool_worker_init (thread_pool_worker_p w, thread_pool_p tp)
{
  w->pool   = tp;
  w->status = THREAD_POOL_WORKER_IDLE;
  
  pthread_mutex_init (&w->mutex, NULL);
  pthread_cond_init (&w->cond, NULL);
  
  if (pthread_create (&w->thread,
		      NULL,
		      thread_pool_worker_function,
		      w))
    w->status = THREAD_POOL_WORKER_ERROR;
}

thread_pool_p
thread_pool_create (size_t num_workers)
{
  thread_pool_p tp;
  int i;
  
  tp = malloc (sizeof (thread_pool_t));
  if (tp)
    {
      tp->num_working = 0;
      tp->num_workers = num_workers;
      tp->workers = malloc (sizeof (thread_pool_worker_t) * num_workers);
      if (!tp->workers)
	goto error;
      
      pthread_mutex_init (&tp->mutex, NULL);
      pthread_cond_init (&tp->cond, NULL);
      
      for (i = 0; i < num_workers; ++i)
	{
	  thread_pool_worker_init (&tp->workers[i], tp);
	}
    }

  return tp;

 error:
  free (tp);
  return NULL;
}

thread_pool_task_id_t
thread_pool_execute (thread_pool_p tp,
		     void (*func) (void*),
		     void* data)
{
  thread_pool_task_id_t id, i;
  thread_pool_worker_p w;
  int num_error_or_stop = 0;
  
  id = -1;
  pthread_mutex_lock (&tp->mutex);
  while (1)
    {
      num_error_or_stop = 0;
      for (i = 0; i < tp->num_workers && id < 0; ++i)
	{
	  w = &tp->workers [i]; 
	  pthread_mutex_lock (&w->mutex);
	  if (w->status == THREAD_POOL_WORKER_IDLE)
	    {
	      w->status = THREAD_POOL_WORKER_WORKING;
	      w->task_func = func;
	      w->task_data = data;
	      pthread_cond_signal (&w->cond);
	      id = i;
	    }
	  else if (w->status != THREAD_POOL_WORKER_WORKING)
	    {
	      num_error_or_stop ++;
	    }
	  pthread_mutex_unlock (&w->mutex);
	}
      
      if (id < 0)
	{
	  pthread_cond_wait (&tp->cond, &tp->mutex);
	}
      else if (num_error_or_stop != tp->num_workers)
	break;
    }
  pthread_mutex_unlock (&tp->mutex);
  return id;
}

void
thread_pool_join (thread_pool_p tp)
{
  thread_pool_worker_p w;
  bool must_join = false;
  int i = 0;
  
  for (i = 0; i < tp->num_workers; ++i)
    {
      w = &tp->workers[i];

      pthread_mutex_lock (&w->mutex);
      must_join =
	w->status == THREAD_POOL_WORKER_IDLE ||
	w->status == THREAD_POOL_WORKER_WORKING;
     
      if (must_join)
	{
	  w->status = THREAD_POOL_WORKER_STOP;
	  pthread_cond_signal (&w->cond);
	}
      pthread_mutex_unlock (&w->mutex);

      if (must_join)
	{
	  pthread_join (w->thread, NULL);
	}
    }
}

static void
thread_pool_worker_release (thread_pool_worker_p w)
{
  pthread_mutex_destroy (&w->mutex);
  pthread_cond_destroy (&w->cond);
}

void
thread_pool_free (thread_pool_p tp)
{
  int i;

  //thread_pool_join (tp); FIXME
  for (i = 0; i < tp->num_workers; ++i) {
    pthread_detach (tp->workers [i].thread);
    thread_pool_worker_release (&tp->workers [i]);
  }
  pthread_cond_destroy (&tp->cond);
  free (tp->workers);
  free (tp);
}
